import RenderInfo from "./util/RenderInfo.js";
import Transform from "./Transform.js";

export default class GameObject {
    constructor(name=null) {
        this.name = name;
        this.transform = new Transform();
        this.components = [];
        this.children = [];
        this.active = true;
    }

    /**
     * Initialize this GameObject, all its components and the all its children recursively
     * 
     * @param {*} gl The WebGL context
     */
    init(gl) {
        this.components.forEach((comp) => {
            if (comp.init)
                comp.init(gl);
        });
        this.children.forEach((child) => {
            if (child.init)
                child.init(gl);
        });
    }

    /**
     * Update the GameObject, all its components and children recursively
     * 
     * @param {Number} deltaTime The delta time since the last call to update
     */
    update(deltaTime) {
        if (!this.active)
            return;
        
        this.components.forEach((comp) => {
            if (comp.update)
                comp.update(deltaTime);
        });
        this.children.forEach((child) => {
            if (child.update)
                child.update(deltaTime);
        });
    }

    /**
     * Render the game object, all its components and children recursively
     * 
     * @param {RenderInfo} renderInfo Render info context
     */
    render(renderInfo) {
        if (!this.active)
            return;
        
        // save the current model view matrix state and apply own transform
        renderInfo.modelViewStack.save();
        renderInfo.modelViewStack.multiply(this.transform.getMat4());

        this.components.forEach((comp) => {
            if (comp.render) {
                comp.render(renderInfo);
            }
        });

        this.children.forEach((child) => {
            if (child.render) {
                child.render(renderInfo);
            }
        });
        
        renderInfo.modelViewStack.restore();
    }

    addComponent(component) {
        this.components.push(component);
        component.gameObject = this;
    }

    removeComponent(component) {
        const idx = this.components.indexOf(component);

        if (idx !== -1) {
            const component = this.components.splice(idx);
            component.gameObject = null;
        }
    }

    addChild(child) {
        this.children.push(child);
        child.gameObject = this;
    }

    removeChild(child) {
        const idx = this.children.indexOf(child);

        if (idx !== -1) {
            const child = this.children.splice(idx);
            child.gameObject = null;
        }
    }

    setActive(active) {
        this.active = active;
        console.log(`${this.name} setActive(${active})`);
    }
}