import MatrixStack from "./util/MatrixStack.js";
import RenderInfo from "./util/RenderInfo.js";
import GameObject from "./GameObject.js";
import CubeRenderer from "./renderers/CubeRenderer.js";
import CylinderRenderer from "./renderers/CylinderRenderer.js";
import Rotator from "./util/Rotator.js";
import PlaneRenderer from "./renderers/PlaneRenderer.js";
import SphereRenderer from "./renderers/SphereRenderer.js";
import HalfSphereRenderer from "./renderers/HalfSphereRenderer.js";

const CAMERA_POSITION = [0, 5, -5];

function main() {
    var canvas = document.querySelector("#glcanvas");
    var gl = null;
    var shaderProgram = null;
    var vr_display = null;
    var vr_present_button = null;
    var programInfo = null;
    var gameObjects = [];


    // Vertex shader program
    const vsSource = `
        attribute vec4 aVertexPosition;
        attribute vec4 aVertexColor; 
        uniform mat4 uModelViewMatrix;
        uniform mat4 uProjectionMatrix;
        varying lowp vec4 vColor;
        void main(void) {
            gl_Position =
                uProjectionMatrix * uModelViewMatrix * aVertexPosition;
            vColor = aVertexColor;
        }
    `;
    // Fragment shader program
    const fsSource = `
        varying lowp vec4 vColor;
        void main(void) {
            gl_FragColor = vColor;
        }
    `;

    function init(preserve_drawing_buffer) {
        var gl_attribs = {
            alpha: false,
            antialias: false,
            preserveDrawingBuffer: preserve_drawing_buffer
        };
        gl = canvas.getContext("webgl", gl_attribs);
        canvas.getContext("webgl", gl_attribs) ||
            canvas.getContext("experimental-webgl", gl_attribs);
        // If we don't have a GL context, give up now
        if (!gl) {
            alert(
                "Unable to initialize WebGL. Your browser or machine " +
                "may not support it."
            );
            return;
        }
        shaderProgram = initShaderProgram(gl, vsSource, fsSource);
        programInfo = {
            program: shaderProgram,
            attribLocations: {
                vertexPosition: gl.getAttribLocation(shaderProgram, "aVertexPosition"),
                vertexColor: gl.getAttribLocation(shaderProgram, "aVertexColor")
            },
            uniformLocations: {
                projectionMatrix: gl.getUniformLocation(
                    shaderProgram,
                    "uProjectionMatrix"
                ),
                modelViewMatrix: gl.getUniformLocation(
                    shaderProgram,
                    "uModelViewMatrix"
                )
            }
        };
        initScene(gl, gameObjects);
        // Wait until we have a WebGL context to resize and start rendering.
        window.addEventListener("resize", onResize, false);
        onResize();
        window.requestAnimationFrame(update);
    }

    function onResize() {
        if (vr_display && vr_display.isPresenting) {
            var left_eye = vr_display.getEyeParameters("left");
            var right_eye = vr_display.getEyeParameters("right");
            canvas.width = Math.max(left_eye.renderWidth, right_eye.renderWidth) * 2;
            canvas.height = Math.max(left_eye.renderHeight, right_eye.renderHeight);
        } else {
            canvas.width = canvas.offsetWidth;
            canvas.height = canvas.offsetHeight;
        }
    }
    var then = 0;
    // Draw the scene repeatedly
    function onContextLost(event) {
        event.preventDefault();
        console.log("WebGL Context Lost.");
        gl = null;
    }

    function onContextRestored(event) {
        console.log("WebGL Context Restored.");
        init(vr_display ? vr_display.capabilities.hasExternalDisplay : false);
    }

    function onVRRequestPresent() {
        vr_display
            .requestPresent([{
                source: canvas
            }])
            .then(
                function () {
                    console.log("hi there!! presenting");
                    // start the sun animation by activating the sun
                    gameObjects.find((go) => go.name == "Sun").setActive(true);
                },
                function (err) {
                    var errMsg = "requestPresent failed.";
                    if (err && err.message) {
                        errMsg += "<br />" + err.message;
                    }
                    console.log(errMsg);
                }
            );
    }

    function onVRExitPresent() {
        if (!vr_display.isPresenting) {
            return;
        }
        vr_display.exitPresent().then(
            function () {
                window.requestAnimationFrame(update);
            },
            function () {
                console.log("exitPresent failed.");
            }
        );
    }

    function onVRPresentChange() {
        onResize();
        var tl = document.getElementById("tl");
        if (vr_display.isPresenting) {
            if (vr_display.capabilities.hasExternalDisplay) {
                removeButton(vr_present_button);
                vr_present_button = addButton("Exit VR", onVRExitPresent);
            }
        } else {
            if (vr_display.capabilities.hasExternalDisplay) {
                removeButton(vr_present_button);
                vr_present_button = addButton("Enter VR", onVRRequestPresent);
            }
        }
    }

    canvas.addEventListener("webglcontextlost", onContextLost, false);
    canvas.addEventListener("webglcontextrestored", onContextRestored, false);
    if (navigator.getVRDisplays) {
        navigator.getVRDisplays().then(
            function (displays) {
                if (displays.length > 0) {
                    vr_display = displays[displays.length - 1];
                    init(true);
                    if (vr_display.capabilities.canPresent) {
                        vr_present_button = addButton("Enter VR", onVRRequestPresent);
                    }
                    window.addEventListener(
                        "vrdisplaypresentchange",
                        onVRPresentChange,
                        false
                    );
                    window.addEventListener(
                        "vrdisplayactivate",
                        onVRRequestPresent,
                        false
                    );
                    window.addEventListener(
                        "vrdisplaydeactivate",
                        onVRExitPresent,
                        false
                    );
                } else {
                    init(false);
                    window.requestAnimationFrame(update);
                    console.log("No VRDisplays found.");
                }
            },
            function () {
                init(false);
                window.requestAnimationFrame(update);
                console.log("No WebVR support.");
            }
        );
    }

    function update(now) {
        now *= 0.001; // convert to seconds
        const deltaTime = now - then;
        then = now;

        gameObjects.forEach((go) => go.update(deltaTime));
        drawScene(vr_display, gl, programInfo, gameObjects);

        if (vr_display && vr_display.isPresenting) {
            vr_display.requestAnimationFrame(update);
        } else {
            window.requestAnimationFrame(update);
        }
    }
}

/*
   Get an orientation from a pose
 */
function getPoseMatrix(out, pose) {
    var orientation = pose.orientation;
    if (!orientation) {
        orientation = [0, 0, 0, 1];
    }
    glMatrix.mat4.fromQuat(out, orientation);
    glMatrix.mat4.invert(out, out);
}

/*
    Init the scene.
*/
function initScene(gl, gameObjects) {
    const sun = new GameObject();
    sun.addComponent(new SphereRenderer(0.5, [1, 221 / 255, 117 / 255, 1], 50, 25));
    sun.transform.setPosition(4, 0, 0);
    
    const sunAnchor = new GameObject("Sun");
    sunAnchor.setActive(false);
    sunAnchor.addComponent(new Rotator({rz: 180 / 20}));
    sunAnchor.addChild(sun);
    gameObjects.push(sunAnchor);

    // set up the smurf house
    // cyclinder for the base
    const houseBody = new GameObject();
    houseBody.addComponent(new CylinderRenderer(1, 0.75, 100, [1, 0.99, 0.8, 1.0]));
    
    // half sphere for the roof
    const houseRoof = new GameObject();
    houseRoof.addComponent(new HalfSphereRenderer(1, [254 / 255, 129 / 255, 109 / 255, 1], 100, 100));
    houseRoof.transform.setPosition(0, 1, 0);
    
    // empty gameobject for the house. attach the childen to it
    const house = new GameObject();
    // house.addComponent(new Rotator({rz: 180 / 20}));
    house.addChild(houseBody);
    house.addChild(houseRoof);
    gameObjects.push(house);

    const grass = new GameObject();
    grass.addComponent(new PlaneRenderer(100, 100, [160 / 255, 243 / 255, 196 / 255, 1]));
    grass.transform.setPosition(0, 0, 0);
    gameObjects.push(grass);

    gameObjects.forEach((go) => go.init(gl));
}
/*
  Draw the scene.
 */
function drawScene(vr_display, gl, programInfo, gameObjects) {
    var canvas = document.querySelector('#glcanvas');
    gl.clearColor(69 / 255, 180 / 255, 231 / 255, 1.0); // Clear to black, fully opaque
    gl.clearDepth(1.0); // Clear everything
    gl.enable(gl.DEPTH_TEST); // Enable depth testing
    gl.depthFunc(gl.LEQUAL); // Near things obscure far things
    // Clear the canvas before we start drawing on it.
    gl.clear(gl.COLOR_BUFFER_BIT | gl.DEPTH_BUFFER_BIT);
    /* Create a perspective matrix, a special matrix that is
       used to simulate the distortion of perspective in a camera.
       Our field of view is 45 degrees, with a width/height
       ratio that matches the display size of the canvas
       and we only want to see objects between 0.1 units
       and 100 units away from the camera.
     */
    const fieldOfView = 45 * Math.PI / 180; // in radians
    const aspect = gl.canvas.clientWidth / gl.canvas.clientHeight;
    const zNear = 0.1;
    const zFar = 100.0;
    if (vr_display) {
        vr_display.depthNear = zNear;
        vr_display.depthFar = zFar;
    }
    const projectionMatrix = glMatrix.mat4.create();
    /* note: glmatrix.js always has the first argument
       as the destination to receive the result. */
    glMatrix.mat4.perspective(projectionMatrix, fieldOfView, aspect,
        zNear, zFar);

    const modelViewStack = new MatrixStack();
    const renderInfo = new RenderInfo(gl, programInfo, modelViewStack);

    // set camera position and rotation
    // create camera matrix, invert it and multiply it to out modelview matrix stack
    const cameraMatrix = glMatrix.mat4.create();
    glMatrix.mat4.targetTo(cameraMatrix, CAMERA_POSITION, [0, 0, 0], [0, 1, 0]);
    const invCameraMatrix = glMatrix.mat4.create();
    glMatrix.mat4.invert(invCameraMatrix, cameraMatrix);

    // render the scene once or twice depending on if we have a vr headset or a normal browser
    if (vr_display && vr_display.isPresenting) {
        const frameData = new VRFrameData();
        vr_display.getFrameData(frameData);
        var poseMatrix = glMatrix.mat4.create();

        // apply the pose to the model view stack
        getPoseMatrix(poseMatrix, frameData.pose);
        modelViewStack.save();
        // apply head pose matrix
        modelViewStack.multiply(poseMatrix);
        // move camera to camera position
        modelViewStack.multiply(invCameraMatrix);

        renderInfo.projectionMatrix = frameData.leftProjectionMatrix;
        // render all game objects for left eye
        gl.viewport(0, 0, canvas.width * 0.5, canvas.height);
        gameObjects.forEach((go) => go.render(renderInfo));

        renderInfo.projectionMatrix = frameData.rightProjectionMatrix;
        // render all game objects for right eye
        gl.viewport(canvas.width * 0.5, 0, canvas.width * 0.5, canvas.height);
        gameObjects.forEach((go) => go.render(renderInfo));

        // restore model view matrix
        modelViewStack.restore();

        vr_display.submitFrame();
    } else {
        renderInfo.projectionMatrix = projectionMatrix;
        // move camera to camera position
        modelViewStack.multiply(invCameraMatrix);

        gameObjects.forEach((go) => go.render(renderInfo));
    }
}

/*
  Initialize a shader program, so WebGL knows how to draw our data
*/
function initShaderProgram(gl, vsSource, fsSource) {
    const vertexShader = loadShader(gl, gl.VERTEX_SHADER, vsSource);
    const fragmentShader = loadShader(gl, gl.FRAGMENT_SHADER, fsSource);
    // Create the shader program
    const shaderProgram = gl.createProgram();
    gl.attachShader(shaderProgram, vertexShader);
    gl.attachShader(shaderProgram, fragmentShader);
    gl.linkProgram(shaderProgram);
    // If creating the shader program failed, alert
    if (!gl.getProgramParameter(shaderProgram, gl.LINK_STATUS)) {
        alert('Unable to initialize the shader program: ' +
            gl.getProgramInfoLog(shaderProgram));
        return null;
    }
    return shaderProgram;
}
/*
  creates a shader of the given type, uploads the source and
  compiles it.
*/
function loadShader(gl, type, source) {
    const shader = gl.createShader(type);
    // Send the source to the shader object
    gl.shaderSource(shader, source);
    // Compile the shader program
    gl.compileShader(shader);
    // See if it compiled successfully
    if (!gl.getShaderParameter(shader, gl.COMPILE_STATUS)) {
        alert('An error occurred compiling the shaders: ' +
            gl.getShaderInfoLog(shader));
        gl.deleteShader(shader);
        return null;
    }
    return shader;
}

function addButtonElement(message) {
    var button_elt = document.createElement("div");
    var webgl_canvas = document.getElementById('glcanvas');
    var button_container = document.getElementById("vr-button-container");
    if (!button_container) {
        button_container = document.createElement("div");
        button_container.id = "vr-button-container";
        Object.assign(button_container.style, {
            fontFamily: "sans-serif",
            position: "absolute",
            zIndex: "999",
            left: "0",
            bottom: "0",
            right: "0",
            margin: "0",
            padding: "0"
        });
        button_container.align = "right";
        webgl_canvas.parentNode.appendChild(button_container);
    }
    Object.assign(button_elt.style, {
        color: "#FFF",
        fontWeight: "bold",
        backgroundColor: "#888",
        borderRadius: "5px",
        border: "3px solid #555",
        position: "relative",
        display: "inline-block",
        margin: "0.5em",
        padding: "0.75em",
        cursor: "pointer",
    });
    button_elt.align = "center";
    button_elt.innerHTML = message;
    button_container.appendChild(button_elt);
    return button_elt;
}

function addButton(message, callback) {
    var element = addButtonElement(message);
    element.addEventListener("click", function (event) {
        callback(event);
        event.preventDefault();
    }, false);
    return {
        element: element,
    };
}

function removeButton(button) {
    if (!button) {
        return;
    }
    if (button.element.parentElement) {
        button.element.parentElement.removeChild(button.element);
    }
}

// register event listener to start the main function
window.addEventListener("load", main, false);