import Component from "../Component.js";

export default class PrimitiveRenderer extends Component{
    constructor() {
        super();

        this.vertices = [];
        this.indices = [];
        this.colors = [];

        this.buffers = {
            vertices: null,
            indices: null,
            colors: null,
        }
    }

    /**
     * Initialize the renderer. Initializes and binds the WebGL buffers for verts, indices and colors
     * 
     * @param {*} gl WebGL context
     */
    init(gl) {
        // create and bind vertex buffer
        this.buffers.vertices = gl.createBuffer();
        gl.bindBuffer(gl.ARRAY_BUFFER, this.buffers.vertices);
        gl.bufferData(gl.ARRAY_BUFFER, new Float32Array(this.vertices), gl.STATIC_DRAW);

        // create and bind index buffer
        this.buffers.indices = gl.createBuffer();
        gl.bindBuffer(gl.ELEMENT_ARRAY_BUFFER, this.buffers.indices);
        gl.bufferData(gl.ELEMENT_ARRAY_BUFFER, new Uint16Array(this.indices), gl.STATIC_DRAW);

        // create and bind color buffer
        this.buffers.colors = gl.createBuffer();
        gl.bindBuffer(gl.ARRAY_BUFFER, this.buffers.colors);
        gl.bufferData(gl.ARRAY_BUFFER, new Float32Array(this.colors), gl.STATIC_DRAW);
    }

    render(renderInfo) {
        const gl = renderInfo.gl;
        const programInfo = renderInfo.programInfo;
        
        /* Tell WebGL how to pull out the positions from the position
           buffer into the vertexPosition attribute
         */
        {
            const numComponents = 3;
            const type = gl.FLOAT;
            const normalize = false;
            const stride = 0;
            const offset = 0;
            gl.bindBuffer(gl.ARRAY_BUFFER, this.buffers.vertices);
            gl.vertexAttribPointer(programInfo.attribLocations.vertexPosition,
                numComponents, type, normalize, stride, offset);
            gl.enableVertexAttribArray(
                programInfo.attribLocations.vertexPosition);
        }
        /* Tell WebGL how to pull out the colors from the color buffer
           into the vertexColor attribute.
        */
        {
            const numComponents = 4;
            const type = gl.FLOAT;
            const normalize = false;
            const stride = 0;
            const offset = 0;
            gl.bindBuffer(gl.ARRAY_BUFFER, this.buffers.colors);
            gl.vertexAttribPointer(programInfo.attribLocations.vertexColor,
                numComponents, type, normalize, stride, offset);
            gl.enableVertexAttribArray(
                programInfo.attribLocations.vertexColor);
        }
        // Tell WebGL which indices to use to index the vertices
        gl.bindBuffer(gl.ELEMENT_ARRAY_BUFFER, this.buffers.indices);
        // Tell WebGL to use our program when drawing
        gl.useProgram(programInfo.program);
    
        // set the projection matrix for the left eye
        gl.uniformMatrix4fv(programInfo.uniformLocations.modelViewMatrix,
            false, renderInfo.modelViewStack.getCurrentMatrix());
        gl.uniformMatrix4fv(programInfo.uniformLocations.projectionMatrix,
            false, renderInfo.projectionMatrix);
        
        // draw all tris
        gl.drawElements(gl.TRIANGLES, this.indices.length, gl.UNSIGNED_SHORT, 0);
    }
}